import 'dart:mirrors';
import 'dart:convert';

class Exportable {
	static bool _isExportableClass(Type type) {
		return reflectClass(type).superclass.reflectedType == Exportable;
	}

	static Type _getActualVariableType(VariableMirror variableMirror) {
		if (variableMirror.type.qualifiedName == reflectType(List).qualifiedName) {
			return variableMirror.type.typeArguments[0].reflectedType;
		}
		return variableMirror.type.reflectedType;
	}

	static bool _isJsonSupported(value) {
		if (value == null || value is bool || value is num || value is String || value is List || value is Map) {
			return true;
		}
		return false;
	}

	static dynamic _importSimpleValue(Type type, value) {
		if (type == DateTime && value is String) {
			return DateTime.parse(value).toLocal();
		} else if (_isJsonSupported(value)) {
			return value;
		}
		return null;
	}

	static Object getInstatiatedTypeForJson(Type type, json, [bool areEqual(String key, String variableName) = null]) {
		if (areEqual == null) areEqual = (key, name) => key == name;

		if (json is String) json = JSON.decode(json);
		if (json == null) return null;
		if (json is List) {
			var result = [];
			json.forEach((e) => result.add(getInstatiatedTypeForJson(type, e)));
			return result;
		} else {

			var result = reflectClass(type).newInstance(const Symbol(''), []).reflectee;
			InstanceMirror thisMirror = reflect(result);
			for (DeclarationMirror declarationMirror in thisMirror.type.declarations.values) {
				if (declarationMirror is VariableMirror) {
					var name = MirrorSystem.getName(declarationMirror.simpleName);

					var actualVariableType = _getActualVariableType(declarationMirror);

					var matchedKeys = json.keys.where((String key) => areEqual(key, name));
					if (matchedKeys.length > 0) {
						var value = json[matchedKeys.elementAt(0)];
						if (_isExportableClass(actualVariableType)) {
							thisMirror.setField(declarationMirror.simpleName, getInstatiatedTypeForJson(actualVariableType, value, areEqual));
						} else {
							thisMirror.setField(declarationMirror.simpleName, _importSimpleValue(declarationMirror.type.reflectedType, value));
						}
					}
				}
			}
			return result;
		}
	}

	static String getJsonFromObject(Object object,[String getName(String name) = null]) {
		if(getName == null) getName = (name) => name;
		return JSON.encode(_getJsonFromObject(object,getName));
	}

	static Object _getJsonFromObject(Object object,String getName(String name)) {
		if (object is List) {
			var list = [];
			for (var obj in object) {
				list.add(_getJsonFromObject(obj,getName));
			}
			return list;
		} else {
			var mapResult = {};
			InstanceMirror thisMirror = reflect(object);
			for (DeclarationMirror declarationMirror in thisMirror.type.declarations.values) {
				if (declarationMirror is VariableMirror) {
					var name = getName(MirrorSystem.getName(declarationMirror.simpleName));
					var value = thisMirror.getField(declarationMirror.simpleName).reflectee;
					if(value is List)
						mapResult[name] = value.map((val) => _getJsonFromObject(val,getName)).toList();
					else
						mapResult[name] = _isExportableClass(declarationMirror.type.reflectedType) ? _getJsonFromObject(value,getName) : value;
				}
			}
			return mapResult;
		}
	}
}
